-- ================================================================================== --
-- FUNCTIONS                                                                          --
-- ================================================================================== --

--
-- SQL FUNCTIONS
--

--
--
-- FUNCTION    : getInterval
--
-- DESCRIPTION : Fetches time in such a way as to force the system
--               to employ this as a literal.

CREATE OR REPLACE FUNCTION ethereal.getInterval(interval) RETURNS TIMESTAMP WITHOUT TIME ZONE AS'
  SELECT LOCALTIMESTAMP - $1
 ' LANGUAGE SQL IMMUTABLE STRICT;

--
--
-- FUNCTION    : getInterval_Hour
--
-- DESCRIPTION : Fetches time in such a way as to force the system
--               to employ this as a literal.

CREATE OR REPLACE FUNCTION ethereal.getInterval_Hour(interval) RETURNS TIMESTAMP WITHOUT TIME ZONE AS'
  SELECT DATE_TRUNC(''hour'', LOCALTIMESTAMP - $1)
 ' LANGUAGE SQL IMMUTABLE STRICT;

--
--
-- FUNCTION    : getInterval_Day
--
-- DESCRIPTION : Fetches time in such a way as to force the system
--               to employ this as a literal.

CREATE OR REPLACE FUNCTION ethereal.getInterval_Day(interval) RETURNS TIMESTAMP WITHOUT TIME ZONE AS'
  SELECT DATE_TRUNC(''day'', LOCALTIMESTAMP - $1)
 ' LANGUAGE SQL IMMUTABLE STRICT;


--
-- PL/PGSQL FUNCTIONS
--

--
--
-- FUNCTION    : getDate
--
-- DESCRIPTION : This will return a properly formatted Date stamp for use
--               in the outside world.

CREATE OR REPLACE FUNCTION ethereal.getDate(TIMESTAMP) RETURNS VARCHAR AS'
  DECLARE
    varTimestamp ALIAS FOR $1;				-- Timestamp

  BEGIN

    -- Quick and simple
    RETURN TO_CHAR(varTimestamp, ''DD-Mon-YYYY'');

  END;
 ' LANGUAGE 'plpgsql';


--
--
-- FUNCTION    : getDate
--
-- DESCRIPTION : This will return a properly formatted Date stamp for use
--               in the outside world.

CREATE OR REPLACE FUNCTION ethereal.getDate(TIMESTAMP WITH TIME ZONE) RETURNS VARCHAR AS'
  DECLARE
    varTimestamp ALIAS FOR $1;				-- Timestamp

  BEGIN

    -- Quick and simple
    RETURN TO_CHAR(varTimestamp, ''DD-Mon-YYYY'');

  END;
 ' LANGUAGE 'plpgsql';


--
--
-- FUNCTION    : getDateTime
--
-- DESCRIPTION : This will return a properly formatted Date/Time stamp for use
--               in the outside world.

CREATE OR REPLACE FUNCTION ethereal.getDateTime(TIMESTAMP) RETURNS VARCHAR AS'
  DECLARE
    varTimestamp ALIAS FOR $1;				-- Timestamp
  BEGIN

    -- Quick and simple
    RETURN TO_CHAR(varTimestamp, ''DD-Mon-YYYY, HH24:MI:SS'');

  END;
 ' LANGUAGE 'plpgsql';


--
--
-- FUNCTION    : getDateTime
--
-- DESCRIPTION : This will return a properly formatted Date/Time stamp for use
--               in the outside world.

CREATE OR REPLACE FUNCTION ethereal.getDateTime(TIMESTAMP WITH TIME ZONE) RETURNS VARCHAR AS'
  DECLARE
    varTimestamp ALIAS FOR $1;				-- Timestamp
  BEGIN

    -- Quick and simple
    RETURN TO_CHAR(varTimestamp, ''DD-Mon-YYYY, HH24:MI:SS'');

  END;
 ' LANGUAGE 'plpgsql';


--
--
-- FUNCTION    : getExpire
--
-- DESCRIPTION : This will return tell us if the timespan for homepages
--               and galleries has in fact expired.  THis is needed
--               with the new date based settings.

CREATE OR REPLACE FUNCTION ethereal.getExpire(VARCHAR) RETURNS VARCHAR AS'
  DECLARE
    -- Alias
    varPuppeteer ALIAS FOR $1;			-- Puppeteer Login

    -- Declarations
    varSpan    TIMESTAMP WITH TIME ZONE;
    varCurrent TIMESTAMP WITH TIME ZONE;

  BEGIN

    -- Pull current time
    varCurrent := ''now'';

    -- Pull expiry for homepage
    SELECT INTO varSpan PuppeteerHomeSpan
         FROM ethereal.Puppeteer
         WHERE LOWER(PuppeteerLogin)=LOWER(varPuppeteer);

    -- Which is greater
    IF varCurrent > varSpan then
      -- Expired
      RETURN ''yes'';
    ELSE
      -- Active
      RETURN ''no'';
    END IF;

  END;
 ' LANGUAGE 'plpgsql';



--
--
-- FUNCTION    : getFile
--
-- DESCRIPTION : This will return a properly formatted Date stamp for use
--               in file names.

CREATE OR REPLACE FUNCTION ethereal.getFile(TIMESTAMP) RETURNS VARCHAR AS'
  DECLARE
    varTimestamp ALIAS FOR $1;				-- Timestamp

  BEGIN

    -- Quick and simple
    RETURN TO_CHAR(varTimestamp, ''YYYYMMDD_HHMM'');

  END;
 ' LANGUAGE 'plpgsql';


--
--
-- FUNCTION    : getFile
--
-- DESCRIPTION : This will return a properly formatted Date stamp for use
--               in filenames

CREATE OR REPLACE FUNCTION ethereal.getFile(TIMESTAMP WITH TIME ZONE) RETURNS VARCHAR AS'
  DECLARE
    varTimestamp ALIAS FOR $1;				-- Timestamp

  BEGIN

    -- Quick and simple
    RETURN TO_CHAR(varTimestamp, ''YYYYMMDD_HHMM'');

  END;
 ' LANGUAGE 'plpgsql';


--
--
-- FUNCTION    : getTime
--
-- DESCRIPTION : This will return a properly formatted Time stamp for use
--               in the outside world.

CREATE OR REPLACE FUNCTION ethereal.getTime(TIMESTAMP) RETURNS VARCHAR AS'
  DECLARE
    varTimestamp ALIAS FOR $1;				-- Timestamp
  BEGIN

    -- Quick and simple
    RETURN TO_CHAR(varTimestamp, ''HH24:MI'');

  END;
 ' LANGUAGE 'plpgsql';


--
--
-- FUNCTION    : getTime
--
-- DESCRIPTION : This will return a properly formatted Time stamp for use
--               in the outside world.

CREATE OR REPLACE FUNCTION ethereal.getTime(TIMESTAMP WITH TIME ZONE) RETURNS VARCHAR AS'
  DECLARE
    varTimestamp ALIAS FOR $1;				-- Timestamp
  BEGIN

    -- Quick and simple
    RETURN TO_CHAR(varTimestamp, ''HH24:MI'');

  END;
 ' LANGUAGE 'plpgsql';

--
--
-- FUNCTION    : getTimestamp
--
-- DESCRIPTION : This will return a timestamp, but will be sure to return which is highest
--               being the current timestamp or whatever was specified

CREATE OR REPLACE FUNCTION ethereal.getTimestamp(TIMESTAMP WITH TIME ZONE) RETURNS TIMESTAMP WITH TIME ZONE AS'
  DECLARE
    varPassed  ALIAS FOR $1;				-- Passed timestamp
    varCurrent TIMESTAMP WITH TIME ZONE;		-- Current timestamp

  BEGIN

    -- Pull current time
    varCurrent := ''now'';

    -- Which is biggest
    IF varCurrent < varPassed THEN
       -- Passed value will do
       RETURN varPassed;

    ELSE
       -- Use current timestamp instead
       RETURN varCurrent;
    END IF;
    
  END;
 ' LANGUAGE 'plpgsql';


--
--
-- FUNCTION    : getTimestamp
--
-- DESCRIPTION : This will return a timestamp, but will be sure to return which is highest
--               being the current timestamp or whatever was specified

CREATE OR REPLACE FUNCTION ethereal.getTimestamp(TIMESTAMP) RETURNS TIMESTAMP AS'
  DECLARE
    varPassed  ALIAS FOR $1;				-- Passed timestamp
    varCurrent TIMESTAMP;		                -- Current timestamp

  BEGIN

    -- Pull current time
    varCurrent := ''now'';

    -- Which is biggest
    IF varCurrent < varPassed THEN
       -- Passed value will do
       RETURN varPassed;

    ELSE
       -- Use current timestamp instead
       RETURN varCurrent;
    END IF;
    
  END;
 ' LANGUAGE 'plpgsql';


--
--
-- FUNCTION    : getPuppet
--
-- DESCRIPTION : This will fetch a default puppet name for the user as it
--               applies.

CREATE OR REPLACE FUNCTION ethereal.getPuppet(VARCHAR, VARCHAR) RETURNS VARCHAR AS'
  DECLARE
    varPuppeteer ALIAS FOR $1;				-- Puppeteer handle
    varPuppet    ALIAS FOR $2;                          -- Puppet name
    varDefault   VARCHAR(30);				-- Puppet default

  BEGIN

    -- Check for a proper default
    SELECT INTO varDefault PuppetName
      FROM ethereal.Puppet
      WHERE PuppetDefault=''yes''
        AND PuppeteerLogin=varPuppeteer
        AND PuppetName <> varPuppet;


    -- No proper default available
    IF varDefault IS NULL THEN

       -- First in the list
       SELECT INTO varDefault PuppetName
         FROM ethereal.Puppet
         WHERE PuppeteerLogin=varPuppeteer
           AND PuppetName <> varPuppet
           AND PuppetPool=''system''
         ORDER BY PuppetName LIMIT 1;

    END IF;


    -- Return default handle
    RETURN varDefault;


  END;
 ' LANGUAGE 'plpgsql';


--
--
-- FUNCTION    : getPuppet
--
-- DESCRIPTION : This will fetch a default puppet name for the user as it
--               applies.

CREATE OR REPLACE FUNCTION ethereal.getPuppet(VARCHAR) RETURNS VARCHAR AS'
  DECLARE
    varPuppeteer ALIAS FOR $1;				-- Puppeteer handle
    varDefault   VARCHAR(30);				-- Puppet default

  BEGIN

    -- Check for a proper default
    SELECT INTO varDefault PuppetName
      FROM ethereal.Puppet
      WHERE PuppetDefault=''yes''
        AND PuppeteerLogin=varPuppeteer;


    -- No proper default available
    IF varDefault IS NULL THEN

       -- First in the list
       SELECT INTO varDefault PuppetName
         FROM ethereal.Puppet
         WHERE PuppeteerLogin=varPuppeteer
           AND PuppetPool=''system''
         ORDER BY PuppetName LIMIT 1;

    END IF;


    -- Return default handle
    RETURN varDefault;


  END;
 ' LANGUAGE 'plpgsql';


--
--
-- FUNCTION    : getWarning
--
-- DESCRIPTION : Fetches the warning level for a specific user/realm combination
--               while applying weight based on the times warnings were made.

CREATE OR REPLACE FUNCTION ethereal.getWarning(VARCHAR, VARCHAR) RETURNS VARCHAR AS'
  DECLARE
    varRealm     ALIAS FOR $1;				-- Puppeteer Login
    varPuppeteer ALIAS FOR $2;				-- Realm Name

    varOne       TIMESTAMP;                             -- Day one
    VarTwo       TIMESTAMP;                             -- Day two
    VarThree     TIMESTAMP;                             -- Day three

    varOccur     RECORD;				-- Occurences seen
    varWeight    NUMERIC(6,2);                          -- Weight calculated

  BEGIN

    -- Pull times
    SELECT INTO varOne   DATE_TRUNC(''day'', CURRENT_TIMESTAMP);
    SELECT INTO varTwo   DATE_TRUNC(''day'', CURRENT_TIMESTAMP - INTERVAL ''1 days'');
    SELECT INTO varThree DATE_TRUNC(''day'', CURRENT_TIMESTAMP - INTERVAL ''2 days'');

    -- Initial assignment
    varWeight := 0;

     -- Loop through results
     FOR VarOccur IN SELECT WarnTimestamp AS When, COUNT(*) AS Times FROM Ethereal.Warning WHERE RealmName=varRealm AND PuppeteerLogin=varPuppeteer GROUP BY WarnTimestamp LOOP
       -- Calculate Weights
       -- Today at full weight
       IF varOccur.When = varOne THEN
          varWeight := varWeight + varOccur.Times;

       -- Yesterday at half weight
       ELSIF varOccur.When = varTwo THEN
          varWeight := varWeight + (varOccur.Times * 0.50);

       -- Day before yesterday at one fourth weight
       ELSIF varOccur.When = varThree THEN
          varWeight := varWeight + (varOccur.Times * 0.25);

       END IF;
    END LOOP;


    -- Speed things up
    varWeight := varWeight * 5;

    -- At most 100%
    IF varWeight > 100 THEN
       -- Set back to 100
       varWeight := 100;
    END IF;


    -- What to return
    IF varWeight = 0 THEN
      -- Empty string
      RETURN '''';
    ELSE
      -- Return warning level
      RETURN varWeight::VARCHAR || ''%'';
    END IF;

  END;
 ' LANGUAGE 'plpgsql';


-- ================================================================================== --
-- TRIGGER/FUNCTIONS                                                                  --
-- ================================================================================== --

--
--
-- TABLE       : Puppet
--
-- OPERATION   : DELETE
--               UPDATE
--
-- DESCRIPTION : Puppet names are generally dragged around to many a location, this
--               trigger will enforce referential integrity where it applies and
--               make sure that puppet names used around are always valid.

CREATE OR REPLACE FUNCTION ethereal.trigPuppet () RETURNS TRIGGER AS '
  DECLARE
    varPuppet VARCHAR(30);				-- Puppet Default

  BEGIN

    --
    -- REMOVE
    IF TG_OP = ''DELETE'' THEN

      -- Removing root handles
      IF OLD.PuppetPool = ''system'' THEN


         -- Cleanup sattelite tables
         -- Ignore table
         EXECUTE ''DELETE FROM ethereal.PuppetIgnore ''
            || '' WHERE PuppetName = ''
            || QUOTE_LITERAL(OLD.PuppetName);


         -- Determine default handle
         SELECT INTO varPuppet * FROM ethereal.getPuppet(OLD.PuppeteerLogin, OLD.PuppetName);



         -- Update tables
         -- Something to replace with
         IF varPuppet IS NOT NULL THEN

            -- Board
            EXECUTE ''UPDATE ethereal.Board SET PuppetName = ''
               || QUOTE_LITERAL(varPuppet)
               || '' WHERE PuppetName = ''
               || QuOTE_LITERAL(OLD.PuppetName);

            -- Comment
            EXECUTE ''UPDATE ethereal.Comment SET PuppetName = ''
               || QUOTE_LITERAL(varPuppet)
               || '' WHERE PuppetName = ''
               || QuOTE_LITERAL(OLD.PuppetName);

            -- Gallery
            EXECUTE ''UPDATE ethereal.Gallery SET GalleryPenName = ''
               || QUOTE_LITERAL(varPuppet)
               || '' WHERE GalleryPenName = ''
               || QuOTE_LITERAL(OLD.PuppetName);

            -- Weblog
            EXECUTE ''UPDATE ethereal.Weblog SET PuppetName = ''
               || QUOTE_LITERAL(varPuppet)
               || '' WHERE PuppetName = ''
               || QuOTE_LITERAL(OLD.PuppetName);


         -- Time to remove
         ELSE

            -- Board
            EXECUTE ''DELETE FROM ethereal.Board ''
               || '' WHERE PuppetName = ''
               || QUOTE_LITERAL(OLD.PuppetName);

            -- Comment
            EXECUTE ''DELETE FROM ethereal.Comment ''
               || '' WHERE PuppetName = ''
               || QUOTE_LITERAL(OLD.PuppetName);

            -- Gallery
            EXECUTE ''DELETE FROM ethereal.Gallery ''
               || '' WHERE GalleryPenName = ''
               || QUOTE_LITERAL(OLD.PuppetName);

            -- Weblog
            EXECUTE ''DELETE FROM ethereal.Weblog ''
               || '' WHERE PuppetName = ''
               || QUOTE_LITERAL(OLD.PuppetName);

         END IF;


      END IF;


      -- Return row unmodified
      RETURN OLD;


    --
    -- UPDATE
    ELSIF TG_OP = ''UPDATE'' THEN

      -- Handle rename
      IF OLD.PuppetName <> NEW.PuppetName AND NEW.PuppetPool = ''system'' THEN

         -- Board
         EXECUTE ''UPDATE ethereal.Board SET PuppetName = ''
            || QUOTE_LITERAL(NEW.PuppetName)
            || '' WHERE PuppetName = ''
            || QuOTE_LITERAL(OLD.PuppetName);

         -- Comment
         EXECUTE ''UPDATE ethereal.Comment SET PuppetName = ''
            || QUOTE_LITERAL(NEW.PuppetName)
            || '' WHERE PuppetName = ''
            || QuOTE_LITERAL(OLD.PuppetName);

         -- Gallery
         EXECUTE ''UPDATE ethereal.Gallery SET GalleryPenName = ''
            || QUOTE_LITERAL(NEW.PuppetName)
            || '' WHERE GalleryPenName = ''
            || QuOTE_LITERAL(OLD.PuppetName);

         -- Puppet Ignore
         EXECUTE ''UPDATE ethereal.PuppetIgnore SET PuppetName = ''
            || QUOTE_LITERAL(NEW.PuppetName)
            || '' WHERE PuppetName = ''
            || QuOTE_LITERAL(OLD.PuppetName);

         -- Weblog
         EXECUTE ''UPDATE ethereal.Weblog SET PuppetName = ''
            || QUOTE_LITERAL(NEW.PuppetName)
            || '' WHERE PuppetName = ''
            || QuOTE_LITERAL(OLD.PuppetName);


      -- Banned handles
      ELSIF NEW.PuppetPool = ''banned'' AND OLD.PuppetPool <> ''banned'' THEN

         -- Board
         EXECUTE ''DELETE FROM ethereal.Board ''
            || '' WHERE PuppetName = ''
            || QUOTE_LITERAL(OLD.PuppetName);

         -- Comment
         EXECUTE ''DELETE FROM ethereal.Comment ''
            || '' WHERE PuppetName = ''
            || QUOTE_LITERAL(OLD.PuppetName);

         -- Gallery
         EXECUTE ''DELETE FROM ethereal.Gallery ''
            || '' WHERE GalleryPenName = ''
            || QUOTE_LITERAL(OLD.PuppetName);

         -- Weblog
         EXECUTE ''DELETE FROM ethereal.Weblog ''
            || '' WHERE PuppetName = ''
            || QUOTE_LITERAL(OLD.PuppetName);
      END IF;

      -- Return row unmodified
      RETURN NEW;

    END IF;


  END;
 ' LANGUAGE 'plpgsql';


--
--
-- TABLE       : Puppeteer
--
-- OPERATION   : DELETE
--               UPDATE
--
-- DESCRIPTION : This trigger will be called based on removals done on the
--               Puppeteer table.   This will result in additional removals
--               within its own table as well as puppets.

CREATE OR REPLACE FUNCTION ethereal.trigPuppeteer () RETURNS TRIGGER AS '
  BEGIN

    --
    -- REMOVE
    IF TG_OP = ''DELETE'' THEN

      -- This is a root handle
      IF OLD.PuppeteerAlias ISNULL THEN

         -- Remove associated handles
         EXECUTE ''DELETE FROM ethereal.Puppeteer WHERE PuppeteerAlias = ''
            || QUOTE_LITERAL(OLD.PuppeteerLogin);

      END IF;

      -- Return row unmodified
      RETURN OLD;


    --
    -- UPDATE
    ELSIF TG_OP = ''UPDATE'' THEN

      -- This user is banned
      IF OLD.PuppeteerBan = ''no'' AND NEW.PuppeteerBan = ''yes'' THEN

         -- Remove canon handles
         EXECUTE ''DELETE FROM ethereal.Puppet WHERE PuppeteerLogin = ''
            || QUOTE_LITERAL(NEW.PuppeteerLogin)
            || '' AND PuppetPool NOT IN (''''system'''', ''''banned'''')'';

         -- Remove duplicate handles
         EXECUTE ''DELETE FROM ethereal.Puppet ''
            || '' WHERE PuppeteerLogin = ''
            || QUOTE_LITERAL(NEW.PuppeteerLogin)
            || '' AND PuppetName IN (''
            || '' SELECT PuppetName FROM ethereal.Puppet WHERE PuppetPool = ''
            || QUOTE_LITERAL(''banned'')
            || '')'';

         -- Transfer remaining handles
         EXECUTE ''UPDATE ethereal.Puppet SET PuppetPool = ''
            || QUOTE_LITERAL(''banned'')
            || '' WHERE PuppeteerLogin = ''
            || QuOTE_LITERAL(NEW.PuppeteerLogin);

      -- Someone has opted to rename
      ELSIF OLD.PuppeteerLogin <> NEW.PuppeteerLogin THEN

         -- Transfer remaining handles
         EXECUTE ''UPDATE ethereal.Puppeteer SET PuppeteerAlias = ''
            || QUOTE_LITERAL(NEW.PuppeteerLogin)
            || '' WHERE PuppeteerAlias = ''
            || QuOTE_LITERAL(OLD.PuppeteerLogin);

      END IF;

      -- Return row unmodified
      RETURN NEW;

    END IF;


  END;
 ' LANGUAGE 'plpgsql';


--
--
-- TABLE       : Realm
--
-- OPERATION   : DELETE
--               INSERT
--               UPDATE
--
-- DESCRIPTION : This trigger will make sure that all associated aliases are
--               enforced properly.  By removing them when appropriate, renaming
--               them or denying an operation.

CREATE OR REPLACE FUNCTION ethereal.trigRealm () RETURNS TRIGGER AS '
  DECLARE
    varRealm VARCHAR(30);				-- Realm existence

  BEGIN

    --
    -- REMOVE
    IF TG_OP = ''DELETE'' THEN

       -- This is a root realm
       IF OLD.RealmAlias IS NULL THEN

         -- Transfer remaining handles
         EXECUTE ''DELETE FROM ethereal.Realm WHERE RealmAlias = ''
            || QUOTE_LITERAL(OLD.RealmName);

       END IF;


      -- Return row unmodified
      RETURN OLD;


    --
    -- INSERT
    ELSIF TG_OP = ''INSERT'' THEN

       -- This is a child realm
       IF NEW.RealmAlias IS NOT NULL THEN

          -- Verification
          SELECT INTO varRealm RealmName FROM ethereal.Realm WHERE RealmName=NEW.RealmAlias;

          -- No match
          IF varRealm IS NULL THEN

             -- Raise a notification
             RAISE NOTICE ''The Realm % does not exist, and cannot be used in an alias.'', NEW.RealmAlias;


             -- Cancel operation
             RETURN NULL;


          END IF;

       END IF;


       -- Just pass it along
       RETURN NEW;


    --
    -- UPDATE
    ELSIF TG_OP = ''UPDATE'' THEN

      -- Someone has opted to rename
      IF OLD.RealmName <> NEW.RealmName THEN

         -- Transfer remaining handles
         EXECUTE ''UPDATE ethereal.Realm SET RealmAlias = ''
            || QUOTE_LITERAL(NEW.RealmName)
            || '' WHERE RealmAlias = ''
            || QuOTE_LITERAL(OLD.RealmName);

      END IF;

      -- Return row unmodified
      RETURN NEW;

    END IF;


  END;
 ' LANGUAGE 'plpgsql';


-- ================================================================================== --
-- TRIGGERS                                                                           --
-- ================================================================================== --

-- CREATE TRIGGER
CREATE TRIGGER trigPuppet
  BEFORE UPDATE OR DELETE
  ON ethereal.Puppet
  FOR EACH ROW EXECUTE PROCEDURE ethereal.trigPuppet();
CREATE TRIGGER trigPuppeteer
  BEFORE UPDATE OR DELETE
  ON ethereal.Puppeteer
  FOR EACH ROW EXECUTE PROCEDURE ethereal.trigPuppeteer();
CREATE TRIGGER trigRealm
  BEFORE UPDATE OR DELETE OR INSERT
  ON ethereal.Realm
  FOR EACH ROW EXECUTE PROCEDURE ethereal.trigRealm();